#include <stdio.h>
#include <string>
#include <fstream>
#include "rasterizer.h"
#include "common.h"
#include "timer.h"
#include "color_space.h"

#include <windows.h>

void create_test_boundary(Array<Line> &lines, const int img_size);

using namespace std;

string filename;
Array2D<double> img;
Array<Line> lines;
int img_width = 64;
int img_height = img_width;

int timing_iters = 0;
bool do_flip = true;


void flip_img(Array2D<double> &img)
{
	for (int j = 0; j < img.s[1]/2; j++)
	{
		for (int i = 0; i < img.s[0]; i++)
		{
			swap(img(i,j), img(i,img.s[1]-1-j));
		}
	}
}

void save_pgm(Array2D<double> &img, string fn, bool flipped = false)
{
	float color_padding = 0.1;

	FILE *f = fopen(fn.c_str(), "wb");
	if (f == 0)
		return;

	fprintf(f, "P5 %d %d 255\n", img.s[0] - FILTER_SIZE-2, img.s[1] - FILTER_SIZE-2);

	int bbx, btx, bby, bty;
	if (!flipped)
	{
		bbx = (int)floor(.5*FILTER_SIZE)+1;
		btx = (int)ceil(.5*FILTER_SIZE)+1;
		bby = (int)ceil(.5*FILTER_SIZE)+1;
		bty = (int)floor(.5*FILTER_SIZE)+1;
	}
	else
	{
		bbx = (int)floor(.5*FILTER_SIZE)+1;
		btx = (int)ceil(.5*FILTER_SIZE)+1;
		bby = (int)floor(.5*FILTER_SIZE)+1;
		bty = (int)ceil(.5*FILTER_SIZE)+1;
	}

	for (int j = bty; j < img.s[1] - bby; j++)
	{
		for (int i = bbx; i < img.s[0] - btx; i++)
		{
			//int val = (int)floor((img(i,j)*(1-2*color_padding) + color_padding)*255 + .5);
			int val = RGB32F_to_sRGB8(img(i,j));
			if (val < 0)
				val = 0;
			else if (val > 255)
				val = 255;
			fprintf(f, "%c", val);
		}
	}

	fclose(f);
}

void save_raw(Array2D<double> &img, string fn, bool flipped = false)
{
	FILE *f = fopen(fn.c_str(), "wb");
	if (f == 0)
		return;

	__int32 w = img.s[0] - FILTER_SIZE-2;
	__int32 h = img.s[1] - FILTER_SIZE-2;
	fwrite((char*)&w, 4, 1, f);
	fwrite((char*)&h, 4, 1, f);

	int bbx, btx, bby, bty;
	if (!flipped)
	{
		bbx = (int)floor(.5*FILTER_SIZE)+1;
		btx = (int)ceil(.5*FILTER_SIZE)+1;
		bby = (int)ceil(.5*FILTER_SIZE)+1;
		bty = (int)floor(.5*FILTER_SIZE)+1;
	}
	else
	{
		bbx = (int)floor(.5*FILTER_SIZE)+1;
		btx = (int)ceil(.5*FILTER_SIZE)+1;
		bby = (int)floor(.5*FILTER_SIZE)+1;
		bty = (int)ceil(.5*FILTER_SIZE)+1;
	}

	for (int j = bty; j < img.s[1] - bby; j++)
	{
		for (int i = bbx; i < img.s[0] - btx; i++)
		{
			float val = img(i,j);
			fwrite((char*)&val, 4, 1, f);
		}
	}

	fclose(f);
}

void read_file()
{
	const double FILTER_OFFSET = -1 + .5*(FILTER_SIZE - 1) - floor(.5*(FILTER_SIZE - 1));

	ifstream f(filename.c_str());
	assert(!f.fail());

	f >> img_width >> img_height;

	int num_lines;
	f >> num_lines;
	lines.resize(num_lines);

	for (int i = 0; i < num_lines; i++)
	{
		Line &c = lines[i];
#ifdef USE_MIXED_CURVES
		f >> c.num_pts;
#else
		int tmp;
		f >> tmp;
#endif
		for (int j = 0; j < c.num_pts; j++)
		{
			f >> c.p[j][0] >> c.p[j][1];
			c.p[j] += FILTER_SIZE + FILTER_OFFSET + 1;
		}
	}

	f.close();
}

int main(int argc, char **argv)
{
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);

	if (argc > 1)
		filename = argv[1];
	else
#ifdef USE_FILE
		filename = "path.txt";
#else
		filename = "";
#endif

	if (argc > 2)
		timing_iters = atoi(argv[2]);

	if (argc > 3)
		do_flip = (bool)atoi(argv[3]);

	// create boundary
	if (filename == "")
		create_test_boundary(lines, img_width);
	else
		read_file();
	
	// rasterize
	img.resize(img_width + FILTER_SIZE + 2 , img_height + FILTER_SIZE + 2);
	for (int i = 0; i < img.data_size; i++)
		img.data[i] = 0;

	rasterize(img, lines);

	// save
	char fn[1024];
	if (filename == "")
	{
		if (do_flip)
			flip_img(img);
		sprintf(fn, "output_perspective_%s.pgm", STR(FILTER_NAME));
		save_pgm(img, fn, do_flip);
	}
	else
	{
		if (do_flip)
			flip_img(img);
		save_pgm(img, "output.pgm", do_flip);
		save_raw(img, "output.raw", do_flip);
	}


	// timings
	if (timing_iters)
	{
		double t1 = get_time();
		for (int i = 0; i < timing_iters; i++)
		{
			rasterize(img, lines);
		}
		double t2 = get_time();
		double t = (t2 - t1) / timing_iters;
		printf("time = %f\n", t*1000);
	}

	return 0;
}
