#include <omp.h>

#include "math.h"
#include "rasterizer.h"
#include "timer.h"
#include "filters.h"

EdgeTable et;

struct ThreadData
{
	bool uninitialized;
	FILTER_NAME filter;
	EdgeTable etp;
	LineStep *ae_root; // edges active for the scan-line
	LineStep *ae_end; // end of edges active for the scan-line
	Scanline sl;

	double *pixel[FILTER_NAME::h];
	double propagate[FILTER_NAME::w][FILTER_NAME::h];

	char spacer[4096]; // make sure that threads are on different pages (pad to 4096 bytes). this makes a huge speed difference.
};

ThreadData tdata[NTHREADS];


///////////////////////////////////////////////////////////////////////////////
// INTERSECTION PARAMETER SOLUTION
///////////////////////////////////////////////////////////////////////////////

#if 0
__forceinline double solve_quad(const double y1, const double h)
{
	return (y1 - sqrt(h - 2*h*y1 + y1*y1))/(2*y1 - 1);
}
#else
// This may be slightly more accurate around y1~=.5, and also a bit faster.
// The issue with y1~=.5 is that both the numerator and the denominator approach zero, and precision is lost.
// In this case, it is best to use a method that is most accurate when the curve is nearly linear.
__forceinline double solve_quad(const double y1, const double h)
{
	// standard solution
	const double e = y1 - .5;
	if (e < .0125 && e > -.0125)
	{
		//double tt = (y1 - sqrt(h - 2*h*y1 + y1*y1))/y2;

#if 0
		// newton iterations are a bit slow
		double t = h;
		for (int i = 0; i < 2; i++)
			t -= (h + t*(-t + 2*(-1 + t)*y1))/(2*(t*y2 - y1));
#else
		// taylor expansions of increasing order 
		// (I have put tolerances required for accuracy to 6 significant figures)
		//double t = h; // |e| < .000002
		//double t = h + 2*e*(-1 + h)*h; // |e| < .002
		double t = h*(1 + 2*e*(-1 + h)*(1 + e*(-2 + 4*h))); // |e| < .0125
		//double t = h*(1 + 2*e*(-1 + h)*(1 + 2*e*(-1 + 2*h + 2*e*(1 + 5*(-1 + h)*h)))); // |e| < .04
#endif

		return t;
	}
	else
	{
		return (y1 - sqrt(h - 2*h*y1 + y1*y1))/(2*e);
	}
}
#endif



double solve_cubic(const double y1, const double y2, const double h)
{
	// Mathc the ends. This works best when there is an inflection.
	//double y1appx = y1*(1-h) + y2*h;
	
	// Best fit quadratic. This works best when there is no inflection.
	double y1appx = (-1 + 3*y1 + 3*y2)*.25; 
	
	double t = solve_quad(y1appx, h);

#if 1
	// use newton iterations. this should converge very quickly wiht a good guess
	double dt = 1;
	while (dt > 1e-6 || dt < -1e-6)
	{
		dt = (-h + t*(3*y1 + t*(t - 6*y1 + 3*t*y1 + 3*y2 - 3*t*y2)))/(3.*(y1 + t*(t - 4*y1 + 3*t*y1 + 2*y2 - 3*t*y2)));
		t -= dt;
	}
#else
	// use newton iterations. this should converge very quickly
	// steps are bounded, with the bound decreasing each step
	// so that in the worst case there is linear convergence.
	// this may be useful when cutting at inflections is
	// disabled, because inflections can cause very bad steps that
	// this prevents.
	double b = .25;
	double dt = 1;
	while (dt > 1e-6 || dt < -1e-6)
	{
		dt = (-h + t*(3*y1 + t*(t - 6*y1 + 3*t*y1 + 3*y2 - 3*t*y2)))/(3.*(y1 + t*(t - 4*y1 + 3*t*y1 + 2*y2 - 3*t*y2)));
		
		double sign = dt < 0 ? -1 : 1;
		if (sign*dt > b)
			t -= sign*b;
		else
			t -= dt;

		b *= .75;
	}
#endif

	return t;
}




///////////////////////////////////////////////////////////////////////////////
// RASTERIZE
///////////////////////////////////////////////////////////////////////////////


void rasterize_line(Array2D<double> &img, int cell_y, ThreadData &td)
{
	//** init line data and check if line exists
	int cell_x = td.sl.end;

	for (int k = 0; k < td.filter.h*td.filter.w; k++)
		((double*)td.propagate)[k] = 0;

	for (int j = 0; j < td.filter.h; j++)
		td.pixel[j] = &img(cell_x, cell_y-j);

	// process each pixel in the scanline
	while (cell_x > td.sl.start - td.filter.w)
	{
		//** add into buffer
		LineRaster *it = td.sl.lists[cell_x];
		while (it)
		{
			td.filter.add_to_vals(td.pixel, td.propagate, *it);
			it = it->next;
		}

		//** fill pixels
		for (int y = 0; y < td.filter.h; y++)
		{
			td.pixel[y]--;
			for (int x = 0; x < td.filter.w; x++)
				td.pixel[y][-x] += td.propagate[x][y];
		}

		//** progress
		td.sl.lists[cell_x] = 0;
		cell_x--;
	}
}

void cut_curve_to_pixels(const int start, const int end, LineStep &c, ThreadData &td)
{
	int curr_x = end;

	//** ADD NEW EDGE
	{
		LineRaster *next = td.sl.lists[curr_x];
		td.sl.heap.s++;
		LineRaster &cut = td.sl.heap[td.sl.heap.s-1];
		td.sl.lists[curr_x] = &cut;
		
		cut.next = next;
#ifdef USE_MIXED_CURVES
		cut.num_pts = c.num_pts;
#endif
		cut.flip_x = c.flip_x;
		cut.flip_y = c.flip_y;

		// set initial increment
		if (c.num_pts == 2)
		{
	#if defined(USE_ONLY_LINES) || defined(USE_MIXED_CURVES)
			cut.px[1][1] = c.py[1-c.flip_x][1];
			cut.px[1][0] = c.py[1-c.flip_x][0] - curr_x;

			if (start == curr_x)
			{
				cut.px[0][1] = c.py[c.flip_x][1];
				cut.px[0][0] = c.py[c.flip_x][0] - curr_x;
			}
			else
			{
				cut.px[0][0] = 0;
				cut.px[0][1] = cut.px[1][1] - cut.px[1][0]*c.line.dy;

				c.px[0] = cut.px[0];
			}
	#endif
		}
		else if (c.num_pts == 3)
		{
	#if defined(USE_ONLY_QUADRATICS) || defined(USE_MIXED_CURVES)
			//** names of xN, yN are mixed, because I am doing the calculations
			//** based on stepping in the y direction. I am trying to reuse code
			//** that is designed for y specifically, rather than making code
			//** that is special cased for the x direction too. x = 1 - y. 
			if (start == curr_x)
			{
				// entire curve is in the line, just copy the whole curve over
				cut.px[0][1] = c.py[2*c.flip_x][1];
				cut.px[0][0] = c.py[2*c.flip_x][0] - curr_x;

				cut.px[1][1] = c.quad.mid_x[1];
				cut.px[1][0] = c.quad.mid_x[0] - curr_x;

				cut.px[2][1] = c.py[2-2*c.flip_x][1];
				cut.px[2][0] = c.py[2-2*c.flip_x][0] - curr_x;
			}
			else
			{
				// scale everything to [0,1] and solve for t
				double scale = c.py[2*c.flip_x][0] - c.py[2-2*c.flip_x][0];

				double y1 = (c.quad.mid_x[0] - c.py[2-2*c.flip_x][0]) / scale;
				double h = (curr_x - c.py[2-2*c.flip_x][0]) / scale;

				double t = solve_quad(y1, h);

				// go to next line
				double x0 = c.py[2-2*c.flip_x][1];
				double x1 = c.quad.mid_x[1];
				double x2 = c.py[2*c.flip_x][1];

				cut.px[2][1] = c.py[2-2*c.flip_x][1];
				cut.px[2][0] = c.py[2-2*c.flip_x][0] - curr_x;

				// get mid point from stepping by t (first half, because t < 1)
				cut.px[1][1] = x0*(1 - t) + x1*t;
				cut.px[1][0] = (c.py[2-2*c.flip_x][0] - curr_x)*(1-t) + (c.quad.mid_x[0] - curr_x)*t;

				// eval end point, obviously y2==1 but x2 is harder
				cut.px[0][1] = x0*(1 - t)*(1 - t) + 2*x1*(1 - t)*t + x2*t*t;
				cut.px[0][0] = 0;

				//** store remainder of curve (really just the mid point is needed, 
				//** because the end is stored in high_y and the start is in py[2])
				c.quad.mid_x[1] = x1*(1 - t) + x2*t;
				c.quad.mid_x[0] = c.quad.mid_x[0]*(1 - t) + c.py[2*c.flip_x][0]*t;

				c.px[0] = cut.px[0];
			}
	#endif
		}
		else// if (c.num_pts == 4)
		{
	#if defined(USE_ONLY_CUBICS) || defined(USE_MIXED_CURVES)
			if (start == curr_x)
			{
				// entire curve is in the line, just copy the whole curve over
				cut.px[0][0] = c.py[3*c.flip_x][0] - curr_x;
				cut.px[0][1] = c.py[3*c.flip_x][1];

				cut.px[1][0] = c.cube.mid_x[c.flip_x][0] - curr_x;
				cut.px[1][1] = c.cube.mid_x[c.flip_x][1];

				cut.px[2][0] = c.cube.mid_x[1-c.flip_x][0] - curr_x;
				cut.px[2][1] = c.cube.mid_x[1-c.flip_x][1];

				cut.px[3][0] = c.py[3-3*c.flip_x][0] - curr_x;
				cut.px[3][1] = c.py[3-3*c.flip_x][1];
			}
			else
			{
				// scale everything to [0,1] and solve for t
				double scale = c.py[3*c.flip_x][0] - c.py[3-3*c.flip_x][0];

				double y1 = (c.cube.mid_x[1-c.flip_x][0] - c.py[3-3*c.flip_x][0]) / scale;
				double y2 = (c.cube.mid_x[c.flip_x][0] - c.py[3-3*c.flip_x][0]) / scale;
				double h = (curr_x - c.py[3-3*c.flip_x][0]) / scale;

				double t = solve_cubic(y1, y2, h);

				// go to next line
				double x0 = c.py[3-3*c.flip_x][1];
				double x1 = c.cube.mid_x[1-c.flip_x][1];
				double x2 = c.cube.mid_x[c.flip_x][1];
				double x3 = c.py[3*c.flip_x][1];

				// get mid point from stepping by t (first half, because t < 1)
				double y0 = c.py[3*c.flip_x][0] - curr_x;
				y1 = c.cube.mid_x[c.flip_x][0] - curr_x;
				y2 = c.cube.mid_x[1-c.flip_x][0] - curr_x;
				double y3 = c.py[3-3*c.flip_x][0] - curr_x;

				cut.px[3][1] = x0;
				cut.px[3][0] = y3;

				cut.px[2][1] = x0*(1-t) + x1*t;
				cut.px[2][0] = y3*(1-t) + y2*t;

				vect2d m;
				m[1] = x1*(1-t) + x2*t;
				m[0] = y2*(1-t) + y1*t;

				cut.px[1][1] = cut.px[2][1]*(1-t) + m[1]*t;
				cut.px[1][0] = cut.px[2][0]*(1-t) + m[0]*t;

				// eval end point, obviously y2==1 but x2 is harder
				cut.px[0][1] = x0*(1-t)*(1-t)*(1-t) + 3*x1*(1-t)*(1-t)*t + 3*x2*(1-t)*t*t + x3*t*t*t;
				cut.px[0][0] = 0;

				//** store remainder of curve (really just the mid point is needed, 
				//** because the end is stored in high_y and the start is in py[2])
				c.cube.mid_x[c.flip_x][1] = x2*(1-t) + x3*t;
				c.cube.mid_x[c.flip_x][0] = c.cube.mid_x[c.flip_x][0]*(1-t) + c.py[3*c.flip_x][0]*t;

				c.cube.mid_x[1-c.flip_x][1] = m[1]*(1-t) + c.cube.mid_x[c.flip_x][1]*t;
				c.cube.mid_x[1-c.flip_x][0] = (m[0]+curr_x)*(1-t) + c.cube.mid_x[c.flip_x][0]*t;

				c.px[0] = cut.px[0];
			}
	#endif
		}

		curr_x--;
	}



	//** UPDATE OLD EDGE
	while (curr_x >= start)
	{
		LineRaster *next = td.sl.lists[curr_x];
		td.sl.heap.s++;
		LineRaster &cut = td.sl.heap[td.sl.heap.s-1];
		td.sl.lists[curr_x] = &cut;
		
		cut.next = next;
#ifdef USE_MIXED_CURVES
		cut.num_pts = c.num_pts;
#endif
		cut.flip_x = c.flip_x;
		cut.flip_y = c.flip_y;

		// update increment from old list
		if (c.num_pts == 2)
		{
#if defined(USE_ONLY_LINES) || defined(USE_MIXED_CURVES)
			cut.px[1][1] = c.px[0][1];
			cut.px[1][0] = 1;

			if (start == curr_x)
			{
				cut.px[0][1] = c.py[c.flip_x][1];
				cut.px[0][0] = c.py[c.flip_x][0] - curr_x;
			}
			else
			{
				cut.px[0][0] = c.px[0][0];
				cut.px[0][1] = c.px[0][1] - c.line.dy;

				c.px[0] = cut.px[0];
			}
#endif
		}
		else if (c.num_pts == 3)
		{
#if defined(USE_ONLY_QUADRATICS) || defined(USE_MIXED_CURVES)
			//** names of xN, yN are mixed, because I am doing the calculations
			//** based on stepping in the y direction. I am trying to reuse code
			//** that is designed for y specifically, rather than making code
			//** that is special cased for the x direction too. x = 1 - y. 
			if (start == curr_x)
			{
				// entire curve is in the line, just copy the whole curve over
				cut.px[2][1] = c.px[0][1];
				cut.px[2][0] = 1;

				cut.px[1][1] = c.quad.mid_x[1];
				cut.px[1][0] = c.quad.mid_x[0] - curr_x;

				cut.px[0][1] = c.py[2*c.flip_x][1];
				cut.px[0][0] = c.py[2*c.flip_x][0] - curr_x;
			}
			else
			{
				// scale everything to [0,1] and solve for t
				double scale = 1 - (c.py[2*c.flip_x][0] - curr_x);

				double y1 = (1 - (c.quad.mid_x[0] - curr_x)) / scale;
				double h = 1 / scale;

				double t = solve_quad(y1, h);

				// go to next line
				double x0 = c.px[0][1];
				double x1 = c.quad.mid_x[1];
				double x2 = c.py[2*c.flip_x][1];

				cut.px[2][1] = c.px[0][1];
				cut.px[2][0] = 1;

				// get mid point from stepping by t (first half, because t < 1)
				cut.px[1][1] = x0*(1 - t) + x1*t;
				cut.px[1][0] = (1-t) + (c.quad.mid_x[0] - curr_x)*t;

				// eval end point, obviously y2==1 but x2 is harder
				cut.px[0][1] = x0*(1 - t)*(1 - t) + 2*x1*(1 - t)*t + x2*t*t;
				cut.px[0][0] = 0;

				//** store remainder of curve (really just the mid point is needed, 
				//** because the end is stored in high_y and the start is in py[2])
				c.quad.mid_x[1] = x1*(1 - t) + x2*t;
				c.quad.mid_x[0] = c.quad.mid_x[0]*(1 - t) + c.py[2*c.flip_x][0]*t;

				c.px[0] = cut.px[0];
			}
#endif
		}
		else// if (c.num_pts == 4)
		{
#if defined(USE_ONLY_CUBICS) || defined(USE_MIXED_CURVES)
			if (start == curr_x)
			{
				// entire curve is in the line, just copy the whole curve over
				cut.px[3][0] = 1;
				cut.px[3][1] = c.px[0][1];

				cut.px[2][0] = c.cube.mid_x[1-c.flip_x][0] - curr_x;
				cut.px[2][1] = c.cube.mid_x[1-c.flip_x][1];

				cut.px[1][0] = c.cube.mid_x[c.flip_x][0] - curr_x;
				cut.px[1][1] = c.cube.mid_x[c.flip_x][1];

				cut.px[0][0] = c.py[3*c.flip_x][0] - curr_x;
				cut.px[0][1] = c.py[3*c.flip_x][1];
			}
			else
			{
				// scale everything to [0,1] and solve for t
				double scale = 1 - (c.py[3*c.flip_x][0] - curr_x);

				double y1 = (1 - (c.cube.mid_x[1-c.flip_x][0] - curr_x)) / scale;
				double y2 = (1 - (c.cube.mid_x[c.flip_x][0] - curr_x)) / scale;
				double h = 1 / scale;

				double t = solve_cubic(y1, y2, h);

				// go to next line
				double x0 = c.px[0][1];
				double x1 = c.cube.mid_x[1-c.flip_x][1];
				double x2 = c.cube.mid_x[c.flip_x][1];
				double x3 = c.py[3*c.flip_x][1];

				// get mid point from stepping by t (first half, because t < 1)
				double y0 = c.py[3*c.flip_x][0] - curr_x;
				y1 = c.cube.mid_x[c.flip_x][0] - curr_x;
				y2 = c.cube.mid_x[1-c.flip_x][0] - curr_x;

				cut.px[3][1] = x0;
				cut.px[3][0] = 1;

				cut.px[2][1] = x0*(1-t) + x1*t;
				cut.px[2][0] = (1-t) + y2*t;

				vect2d m;
				m[1] = x1*(1-t) + x2*t;
				m[0] = y2*(1-t) + y1*t;

				cut.px[1][1] = cut.px[2][1]*(1-t) + m[1]*t;
				cut.px[1][0] = cut.px[2][0]*(1-t) + m[0]*t;

				// eval end point, obviously y2==1 but x2 is harder
				cut.px[0][1] = x0*(1-t)*(1-t)*(1-t) + 3*x1*(1-t)*(1-t)*t + 3*x2*(1-t)*t*t + x3*t*t*t;
				cut.px[0][0] = 0;

				//** store remainder of curve (really just the mid point is needed, 
				//** because the end is stored in high_y and the start is in py[2])
				c.cube.mid_x[c.flip_x][1] = x2*(1-t) + x3*t;
				c.cube.mid_x[c.flip_x][0] = c.cube.mid_x[c.flip_x][0]*(1-t) + c.py[3*c.flip_x][0]*t;

				c.cube.mid_x[1-c.flip_x][1] = m[1]*(1-t) + c.cube.mid_x[c.flip_x][1]*t;
				c.cube.mid_x[1-c.flip_x][0] = (m[0]+curr_x)*(1-t) + c.cube.mid_x[c.flip_x][0]*t;

				c.px[0] = cut.px[0];
			}
#endif
		}

		curr_x--;
	}
}


void update_ae_list(LineStep *a_root, int curr_y, ThreadData &td)
{
	td.sl.heap.s = 0;
	td.sl.start = 2000000000;
	td.sl.end = -2000000000;

	// incrementally update old edges
	LineStep *ita = td.ae_root, *prev_ae = 0;
	while (ita)
	{
		//** FROM OLD LIST
		LineStep &c = *ita;

		if (c.end_y < curr_y)
		{
			// remove from list, because it is past the end
			if (prev_ae == 0 && ita->next == 0)
			{
				td.ae_root = 0;
				break;
			}

			if (prev_ae)
			{
				prev_ae->next = ita->next;
				if (ita->next == 0)
					td.ae_end = prev_ae;
			}
			else
				td.ae_root = ita->next;

			ita = ita->next;
			continue;
		}

		// update increment from old list
		if (c.num_pts == 2)
		{
#if defined(USE_ONLY_LINES) || defined(USE_MIXED_CURVES)
			c.py[0][0] = c.py[1][0];
			c.py[0][1] = 0; // needed because c.py[0][1] may be non-zero on the first increment

			if (c.end_y == curr_y)
			{
				c.py[1][0] = c.high_y[0];
				c.py[1][1] = c.high_y[1] - c.end_y;
			}
			else
			{
				c.py[1][0] += c.line.dx;
			}
#endif
		}
		else if (c.num_pts == 3)
		{
#if defined(USE_ONLY_QUADRATICS) || defined(USE_MIXED_CURVES)
			if (c.end_y == curr_y)
			{
				// entire curve is in the line, just copy the whole curve over
				c.py[0][0] = c.py[2][0];
				c.py[0][1] = 0;

				c.py[1][0] = c.quad.mid_y[0];
				c.py[1][1] = c.quad.mid_y[1] - curr_y;

				c.py[2][0] = c.high_y[0];
				c.py[2][1] = c.high_y[1] - curr_y;

				c.quad.mid_x = c.py[1];
			}
			else
			{
				// scale everything to [0,1] and solve for t
				double scale = c.high_y[1] - curr_y;

				double y1 = (c.quad.mid_y[1] - curr_y) / scale;
				double h = 1 / scale;

				double t = solve_quad(y1, h);
				
				// go to next line
				double x0 = c.py[2][0];
				double x1 = c.quad.mid_y[0];
				double x2 = c.high_y[0];

				vect2d p[3];

				p[0][0] = c.py[2][0];
				p[0][1] = 0;

				// get mid point from stepping by t (first half, because t < 1)
				p[1][0] = x0*(1 - t) + x1*t;
				p[1][1] = (c.quad.mid_y[1] - curr_y)*t;

				// eval end point, obviously y2==1 but x2 is harder
				p[2][0] = x0*(1 - t)*(1 - t) + 2*x1*(1 - t)*t + x2*t*t;
				p[2][1] = 1;

				//** store remainder of curve (really just the mid point is needed, 
				//** because the end is stored in high_y and the start is in py[2])
				c.quad.mid_y[0] = x1*(1 - t) + x2*t;
				c.quad.mid_y[1] = c.quad.mid_y[1]*(1 - t) + c.high_y[1]*t;
				
				c.py[0] = p[0];
				c.py[1] = p[1];
				c.py[2] = p[2];

				c.quad.mid_x = c.py[1];
			}
#endif
		}
		else// if (c.num_pts == 4)
		{
#if defined(USE_ONLY_CUBICS) || defined(USE_MIXED_CURVES)
			if (c.end_y == curr_y)
			{
				// entire curve is in the line, just copy the whole curve over
				c.py[0][0] = c.py[3][0];
				c.py[0][1] = 0;

				c.py[1][0] = c.cube.mid_y[0][0];
				c.py[1][1] = c.cube.mid_y[0][1] - curr_y;

				c.py[2][0] = c.cube.mid_y[1][0];
				c.py[2][1] = c.cube.mid_y[1][1] - curr_y;

				c.py[3][0] = c.high_y[0];
				c.py[3][1] = c.high_y[1] - curr_y;

				c.cube.mid_x[0] = c.py[1];
				c.cube.mid_x[1] = c.py[2];
			}
			else
			{
				// scale everything to [0,1] and solve for t
				double scale = c.high_y[1] - curr_y;

				double y1 = (c.cube.mid_y[0][1] - curr_y) / scale;
				double y2 = (c.cube.mid_y[1][1] - curr_y) / scale;
				double h = 1 / scale;

				double t = solve_cubic(y1, y2, h);
				
				// go to next line
				double y0 = 0;
				y1 = c.cube.mid_y[0][1] - curr_y;
				y2 = c.cube.mid_y[1][1] - curr_y;

				double x0 = c.py[3][0];
				double x1 = c.cube.mid_y[0][0];
				double x2 = c.cube.mid_y[1][0];
				double x3 = c.high_y[0];

				vect2d p[4];

				p[0][0] = x0;
				p[0][1] = 0;

				// get mid point from stepping by t (first half, because t < 1)
				p[1][0] = x0*(1-t) + x1*t;
				p[1][1] = y0*(1-t) + y1*t;

				vect2d m;
				m[0] = x1*(1-t) + x2*t;
				m[1] = y1*(1-t) + y2*t;

				p[2][0] = p[1][0]*(1-t) + m[0]*t;
				p[2][1] = p[1][1]*(1-t) + m[1]*t;

				// eval end point, obviously y2==1 but x2 is harder
				p[3][0] = x0*(1-t)*(1-t)*(1-t) + 3*x1*(1-t)*(1-t)*t + 3*x2*(1-t)*t*t + x3*t*t*t;
				p[3][1] = 1;

				//** store remainder of curve (really just the mid point is needed, 
				//** because the end is stored in high_y and the start is in py[2])
				c.cube.mid_y[1][0] = x2*(1-t) + x3*t;
				c.cube.mid_y[1][1] = c.cube.mid_y[1][1]*(1-t) + c.high_y[1]*t;
				
				c.cube.mid_y[0][0] = m[0]*(1-t) + c.cube.mid_y[1][0]*t;
				c.cube.mid_y[0][1] = (m[1]+curr_y)*(1-t) + c.cube.mid_y[1][1]*t;

				c.py[0] = p[0];
				c.py[1] = p[1];
				c.py[2] = p[2];
				c.py[3] = p[3];
				
				c.cube.mid_x[0] = c.py[1];
				c.cube.mid_x[1] = c.py[2];
			}
#endif
		}

		// set ends
		int start, end;
		if (c.py[0][0] > c.py[c.num_pts-1][0])
		{
			c.flip_x = 1;
			start = (int)c.py[c.num_pts-1][0];
			end = (int)c.py[0][0];
			if (c.py[0][0] == floor(c.py[0][0]))
				end--;
		}
		else
		{
			c.flip_x = 0;
			start = (int)c.py[0][0];
			end = (int)c.py[c.num_pts-1][0];
			if (c.py[c.num_pts-1][0] == floor(c.py[c.num_pts-1][0]))
				end--;
		}
		if (start > end)
			start--;
		td.sl.start = min(td.sl.start, start);
		td.sl.end = max(td.sl.end, end);

		// cut to pixels
		cut_curve_to_pixels(start, end, c, td);

		// update iterator
		prev_ae = ita;
		ita = ita->next;
	}

	// set config of input edges
	LineStep *itb = a_root;
	LineStep *a_end = 0;
	while (itb)
	{
		//** FROM NEW LIST
		a_end = itb;
		LineStep &c = *itb;

		// set initial increment
		if (c.num_pts == 2)
		{
#if defined(USE_ONLY_LINES) || defined(USE_MIXED_CURVES)
			c.py[0][0] = c.py[1][0];
			c.py[0][1] = c.py[1][1] - curr_y;

			if (c.end_y == curr_y)
			{
				c.py[1][0] = c.high_y[0];
				c.py[1][1] = c.high_y[1] - c.end_y;
			}
			else
			{
				c.py[1][1] = 1;
				c.py[1][0] = c.py[0][0] + (1 - c.py[0][1])*c.line.dx;
			}
#endif
		}
		else if (c.num_pts == 3)
		{
#if defined(USE_ONLY_QUADRATICS) || defined(USE_MIXED_CURVES)
			if (c.end_y == curr_y)
			{
				// entire curve is in the line, just copy the whole curve over
				c.py[0][0] = c.py[2][0];
				c.py[0][1] = c.py[2][1] - curr_y;

				c.py[1][0] = c.quad.mid_y[0];
				c.py[1][1] = c.quad.mid_y[1] - curr_y;

				c.py[2][0] = c.high_y[0];
				c.py[2][1] = c.high_y[1] - curr_y;
				
				c.quad.mid_x = c.py[1];
			}
			else
			{
				// scale everything to [0,1] and solve for t
				double scale = c.high_y[1] - c.py[2][1];

				double y1 = (c.quad.mid_y[1] - c.py[2][1]) / scale;
				double h = (1 - (c.py[2][1] - curr_y)) / scale;

				double t = solve_quad(y1, h);
				
				// go to next line
				double x0 = c.py[2][0];
				double x1 = c.quad.mid_y[0];
				double x2 = c.high_y[0];

				vect2d p[3];

				p[0][0] = c.py[2][0];
				p[0][1] = c.py[2][1] - curr_y;

				// get mid point from stepping by t (first half, because t < 1)
				p[1][0] = x0*(1 - t) + x1*t;
				p[1][1] = (c.py[2][1] - curr_y)*(1-t) + (c.quad.mid_y[1] - curr_y)*t;

				// eval end point, obviously y2==1 but x2 is harder
				p[2][0] = x0*(1 - t)*(1 - t) + 2*x1*(1 - t)*t + x2*t*t;
				p[2][1] = 1;

				//** store remainder of curve (really just the mid point is needed, 
				//** because the end is stored in high_y and the start is in py[2])
				c.quad.mid_y[0] = x1*(1 - t) + x2*t;
				c.quad.mid_y[1] = c.quad.mid_y[1]*(1 - t) + c.high_y[1]*t;
				
				c.py[0] = p[0];
				c.py[1] = p[1];
				c.py[2] = p[2];
				
				c.quad.mid_x = c.py[1];
			}
#endif
		}
		else// if (c.num_pts == 4)
		{
#if defined(USE_ONLY_CUBICS) || defined(USE_MIXED_CURVES)
			if (c.end_y == curr_y)
			{
				// entire curve is in the line, just copy the whole curve over
				c.py[0][0] = c.py[3][0];
				c.py[0][1] = c.py[3][1] - curr_y;

				c.py[1][0] = c.cube.mid_y[0][0];
				c.py[1][1] = c.cube.mid_y[0][1] - curr_y;

				c.py[2][0] = c.cube.mid_y[1][0];
				c.py[2][1] = c.cube.mid_y[1][1] - curr_y;

				c.py[3][0] = c.high_y[0];
				c.py[3][1] = c.high_y[1] - curr_y;

				c.cube.mid_x[0] = c.py[1];
				c.cube.mid_x[1] = c.py[2];
			}
			else
			{
				// scale everything to [0,1] and solve for t
				double scale = c.high_y[1] - c.py[3][1];

				double y1 = (c.cube.mid_y[0][1] - c.py[3][1]) / scale;
				double y2 = (c.cube.mid_y[1][1] - c.py[3][1]) / scale;
				double h = (1 - (c.py[3][1] - curr_y)) / scale;

				double t = solve_cubic(y1, y2, h);
				
				// go to next line
				double x0 = c.py[3][0];
				double x1 = c.cube.mid_y[0][0];
				double x2 = c.cube.mid_y[1][0];
				double x3 = c.high_y[0];

				vect2d p[4];

				// get mid point from stepping by t (first half, because t < 1)
				double y0 = c.py[3][1] - curr_y;
				y1 = c.cube.mid_y[0][1] - curr_y;
				y2 = c.cube.mid_y[1][1] - curr_y;

				p[0][0] = x0;
				p[0][1] = y0;

				p[1][0] = x0*(1-t) + x1*t;
				p[1][1] = y0*(1-t) + y1*t;

				vect2d m;
				m[0] = x1*(1-t) + x2*t;
				m[1] = y1*(1-t) + y2*t;

				p[2][0] = p[1][0]*(1-t) + m[0]*t;
				p[2][1] = p[1][1]*(1-t) + m[1]*t;

				// eval end point, obviously y2==1 but x2 is harder
				p[3][0] = x0*(1-t)*(1-t)*(1-t) + 3*x1*(1-t)*(1-t)*t + 3*x2*(1-t)*t*t + x3*t*t*t;
				p[3][1] = 1;
				//p[3][1] = y0*(1-t)*(1-t)*(1-t) + 3*y1*(1-t)*(1-t)*t + 3*y2*(1-t)*t*t + y3*t*t*t;

				//** store remainder of curve (really just the mid point is needed, 
				//** because the end is stored in high_y and the start is in py[2])
				c.cube.mid_y[1][0] = x2*(1-t) + x3*t;
				c.cube.mid_y[1][1] = c.cube.mid_y[1][1]*(1-t) + c.high_y[1]*t;
				
				c.cube.mid_y[0][0] = m[0]*(1-t) + c.cube.mid_y[1][0]*t;
				c.cube.mid_y[0][1] = (m[1]+curr_y)*(1-t) + c.cube.mid_y[1][1]*t;

				c.py[0] = p[0];
				c.py[1] = p[1];
				c.py[2] = p[2];
				c.py[3] = p[3];
				
				c.cube.mid_x[0] = c.py[1];
				c.cube.mid_x[1] = c.py[2];
			}
#endif
		}

		// set ends
		int start, end;
		if (c.py[0][0] > c.py[c.num_pts-1][0])
		{
			c.flip_x = 1;
			start = (int)c.py[c.num_pts-1][0];
			end = (int)c.py[0][0];
			if (c.py[0][0] == floor(c.py[0][0]))
				end--;
		}
		else
		{
			c.flip_x = 0;
			start = (int)c.py[0][0];
			end = (int)c.py[c.num_pts-1][0];
			if (c.py[c.num_pts-1][0] == floor(c.py[c.num_pts-1][0]))
				end--;
		}
		if (start > end)
			start--;
		td.sl.start = min(td.sl.start, start);
		td.sl.end = max(td.sl.end, end);

		// cut to pixels
		cut_curve_to_pixels(start, end, c, td);

		// update iterator
		itb = itb->next;
	}

	//** merge the lists of old edges and new edges
	if (a_root)
	{
		a_end->next = td.ae_root;
		td.ae_root = a_root;
	}
}


void calc_end_points(LineStep *a_root, int curr_y, int curr_y_start, ThreadData &td)
{
	// incrementally update old edges
	LineStep *ita = td.ae_root, *prev_ae = 0;
	while (ita)
	{
		//** FROM OLD LIST
		LineStep &c = *ita;

		if (c.end_y <= curr_y)
		{
			// remove from list, because it is past the end
			if (prev_ae == 0 && ita->next == 0)
			{
				td.ae_root = 0;
				break;
			}
			
			if (prev_ae)
			{
				prev_ae->next = ita->next;
				if (ita->next == 0)
					td.ae_end = prev_ae;
			}
			else
				td.ae_root = ita->next;

			ita = ita->next;
			continue;
		}

		// update increment from old list
		if (c.num_pts == 2)
		{
#if defined(USE_ONLY_LINES) || defined(USE_MIXED_CURVES)
			double py0 = max(c.py[1][1], (double)curr_y_start);

			//if (c.end_y != curr_y)
			{
				c.py[1][0] += c.line.dx * (1 + curr_y - py0);
				c.py[1][1] = 1;
			}
#endif
		}
		else if (c.num_pts == 3)
		{
#if defined(USE_ONLY_QUADRATICS) || defined(USE_MIXED_CURVES)
			//if (c.end_y != curr_y)
			{
				double py0 = max(c.py[2][1], (double)curr_y_start);

				// scale everything to [0,1] and solve for t
				double scale = c.high_y[1] - py0;

				double y1 = (c.quad.mid_y[1] - py0) / scale;
				double h = (1 - (py0 - curr_y)) / scale;

				double t = solve_quad(y1, h);
				
				// go to next line
				double x0 = c.py[2][0];
				double x1 = c.quad.mid_y[0];
				double x2 = c.high_y[0];

				vect2d p[3];

				// eval end point, obviously y2==1 but x2 is harder
				p[2][0] = x0*(1 - t)*(1 - t) + 2*x1*(1 - t)*t + x2*t*t;
				p[2][1] = 1;

				//** store remainder of curve (really just the mid point is needed, 
				//** because the end is stored in high_y and the start is in py[2])
				c.quad.mid_y[0] = x1*(1 - t) + x2*t;
				c.quad.mid_y[1] = c.quad.mid_y[1]*(1 - t) + c.high_y[1]*t;
				
				c.py[2] = p[2];
			}
#endif
		}
		else// if (c.num_pts == 4)
		{
#if defined(USE_ONLY_CUBICS) || defined(USE_MIXED_CURVES)
			//if (c.end_y != curr_y)
			{
				double py0 = max(c.py[3][1], (double)curr_y_start);

				// scale everything to [0,1] and solve for t
				double scale = c.high_y[1] - py0;

				double y0 = py0 - curr_y;
				double y1 = (c.cube.mid_y[0][1] - py0) / scale;
				double y2 = (c.cube.mid_y[1][1] - py0) / scale;
				double h = (1 - y0) / scale;

				double t = solve_cubic(y1, y2, h);
				
				// go to next line
				y1 = c.cube.mid_y[0][1] - curr_y;
				y2 = c.cube.mid_y[1][1] - curr_y;

				double x0 = c.py[3][0];
				double x1 = c.cube.mid_y[0][0];
				double x2 = c.cube.mid_y[1][0];
				double x3 = c.high_y[0];

				vect2d p[4];

				vect2d m;
				m[0] = x1*(1-t) + x2*t;
				m[1] = y1*(1-t) + y2*t;

				// eval end point, obviously y2==1 but x2 is harder
				p[3][0] = x0*(1-t)*(1-t)*(1-t) + 3*x1*(1-t)*(1-t)*t + 3*x2*(1-t)*t*t + x3*t*t*t;
				p[3][1] = 1;

				//** store remainder of curve (really just the mid point is needed, 
				//** because the end is stored in high_y and the start is in py[2])
				c.cube.mid_y[1][0] = x2*(1-t) + x3*t;
				c.cube.mid_y[1][1] = c.cube.mid_y[1][1]*(1-t) + c.high_y[1]*t;
				
				c.cube.mid_y[0][0] = m[0]*(1-t) + c.cube.mid_y[1][0]*t;
				c.cube.mid_y[0][1] = (m[1]+curr_y)*(1-t) + c.cube.mid_y[1][1]*t;

				c.py[3] = p[3];
			}
#endif
		}

		// update iterator
		prev_ae = ita;
		ita = ita->next;
	}
}

__forceinline void skip_to_next_scanline(int iscan, ThreadData &td)
{
#if 1
	// with an end pointer, we don't have to walk the list each scanline.
	if (td.etp.lists[iscan])
	{
		td.etp.list_ends[iscan]->next = td.ae_root;
		td.ae_root = td.etp.lists[iscan];
	}
#else
	// remove old edges
	LineStep *ita = td.ae_root, *prev_ae = 0;
	while (ita)
	{
		LineStep &c = *ita;

		if (c.end_y < curr_y)
		{
			// remove from list, because it is past the end
			if (prev_ae == 0 && ita->next == 0)
			{
				td.ae_root = 0;
				break;
			}
			
			if (prev_ae)
				prev_ae->next = ita->next;
			else
				td.ae_root = ita->next;

			ita = ita->next;
			continue;
		}

		// update iterator
		prev_ae = ita;
		ita = ita->next;
	}

	//** merge the lists of old edges and new edges
	if (prev_ae)
		prev_ae->next = a_root;
	else
		td.ae_root = a_root;
#endif
}

void cut_bez3(Line &c, Line &a, Line &b, double t)
{
	vect2d &a0 = c.p[0];
	vect2d &a1 = c.p[1];
	vect2d &a2 = c.p[2];
	vect2d &a3 = c.p[3];
	
	vect2d &b0 = a.p[1];
	vect2d b1;
	vect2d &b2 = b.p[2];

	vect2d &c0 = a.p[2];
	vect2d &c1 = b.p[1];

	vect2d &d0 = a.p[3];

	double t1 = 1-t;

	b0 = a0*t1 + a1*t;
	b1 = a1*t1 + a2*t;
	b2 = a2*t1 + a3*t;

	c0 = b0*t1 + b1*t;
	c1 = b1*t1 + b2*t;

	d0 = c0*t1 + c1*t;

	b.p[0] = a.p[3];
	a.p[0] = c.p[0];
	b.p[3] = c.p[3];

#ifdef USE_MIXED_CURVES
	a.num_pts = 4;
	b.num_pts = 4;
#endif
}

void cut_bez2(Line &c, Line &a, Line &b, double t)
{
	vect2d &a1 = a.p[1];
	vect2d &b1 = b.p[1];
	vect2d &a2 = a.p[2];
	double t1 = 1-t;

	a.p[0] = c.p[0];
	b.p[2] = c.p[2];

	a1 = c.p[0]*t1 + c.p[1]*t;
	b1 = c.p[1]*t1 + c.p[2]*t;

	a2 = a1*t1 + b1*t;

	b.p[0] = a.p[2];

#ifdef USE_MIXED_CURVES
	a.num_pts = 3;
	b.num_pts = 3;
#endif
}

void subdiv_bez2(Line &c, Array<Line> &cut)
{
#if 0
	// elliptic distance
	double a1 = (c.p[1] - c.p[0]).length();
	double a2 = (c.p[2] - c.p[1]).length();
	double d = (c.p[2] - c.p[0]).length();
	double x = a1 + a2 - d;

	const double tol = .01;
#elif 0
	// distance to line
	vect2d a = c.p[1] - c.p[0];
	vect2d b = c.p[2] - c.p[0];

	double x = ( a - ((a*b)/b.length2())*b ).length2();
	
	const double tol12 = 1;
	const double tol = tol12*tol12;
#else
	// distance to line segment
	vect2d a = c.p[1] - c.p[0];
	vect2d b = c.p[2] - c.p[0];

	//double d = (a*b)/b.length2();
	double d = (a[0]*b[0] + a[1]*b[1])/b.length2();

	double x;
	if (d <= 0)
		x = a.length2();
	else if (d >= 1)
		x = (c.p[1] - c.p[2]).length2();
	else
		x = ( a - d*b ).length2();
	
	const double tol12 = 2 * .01;
	const double tol = tol12*tol12;
#endif

	if (x < tol)
	{
		Line l;
#ifdef USE_MIXED_CURVES
		l.num_pts = 2;
#endif
		l.p[0] = c.p[0];
		l.p[1] = c.p[2];
		cut.push_back(l);
	}
	else
	{
		Line s[2];
		cut_bez2(c, s[0], s[1], .5);
		subdiv_bez2(s[0], cut);
		subdiv_bez2(s[1], cut);
	}
}

void subdiv_bez3(Line &c, Array<Line> &cut)
{
#if 0
	// elliptic distance extension
	double a1 = (c.p[1] - c.p[0]).length();
	double a2 = (c.p[2] - c.p[1]).length();
	double a3 = (c.p[3] - c.p[2]).length();
	double d = (c.p[3] - c.p[0]).length();
	double x = a1 + a2 + a3 - d;

	double tol = .01;
#else
	// distance to line segment, max
	vect2d a1 = c.p[1] - c.p[0];
	vect2d a2 = c.p[2] - c.p[0];
	vect2d b = c.p[3] - c.p[0];

	//double d1 = (a1*b)/b.length2();
	//double d2 = (a2*b)/b.length2();
	double d1 = (a1[0]*b[0] + a1[1]*b[1])/b.length2();
	double d2 = (a2[0]*b[0] + a2[1]*b[1])/b.length2();

	double x1, x2;

	if (d1 <= 0)
		x1 = a1.length();
	else if (d1 >= 1)
		x1 = (c.p[1] - c.p[3]).length();
	else
		x1 = ( a1 - d1*b ).length();
	
	if (d2 <= 0)
		x2 = a2.length();
	else if (d2 >= 1)
		x2 = (c.p[2] - c.p[3]).length();
	else
		x2 = ( a2 - d2*b ).length();

	double x = max(x1, x2);

	const double tol = .35;
#endif

	if (x < tol)
	{
		Line l;
#ifdef USE_MIXED_CURVES
		l.num_pts = 2;
#endif
		l.p[0] = c.p[0];
		l.p[1] = c.p[3];
		cut.push_back(l);
	}
	else
	{
		Line s[2];
		cut_bez3(c, s[0], s[1], .5);
		subdiv_bez3(s[0], cut);
		subdiv_bez3(s[1], cut);
	}
}

void cut_monotonic(Array<Line> &uncut, Array<Line> &cut)
{
#if 0
	//*** CUT INTO LINE SEGMENTS THAT APPROXIMATE THE CURVES
	for (int i = 0; i < uncut.s; i++)
	{
		if (uncut[i].num_pts == 2)
			cut.push_back(uncut[i]);
		else if (uncut[i].num_pts == 3)
			subdiv_bez2(uncut[i], cut);
		else if (uncut[i].num_pts == 4)
			subdiv_bez3(uncut[i], cut);
	}
	//printf("cut into %d lines\n", cut.s);
#else
	//*** CUT INTO MONOTONIC AND INFLECTION FREE PIECES

	//cut.reserve(uncut.s); // for linear
	//cut.reserve(uncut.s*3); // for quadratic
	cut.reserve(uncut.s*7); // for cubic
	Line *cp = cut.data;

	Line *un = uncut.data;
	Line *unending = uncut.data + uncut.s;

	while (un < unending)
	{
		if (un->num_pts == 4)
		{
			static Array<double> t_vals;
			t_vals.s = 0;
			double t;

			//** find all parameter vals to cut at

			// cut in y-direction
			{
				const double y0 = un->p[0][1];
				const double y1 = un->p[1][1];
				const double y2 = un->p[2][1];
				const double y3 = un->p[3][1];

				double a = y0 - 2*y1 + y2;
				double b = y0 - 3*y1 + 3*y2 - y3;
				double c = y1*y1 - y0*y2 - y1*y2 + y2*y2 + y0*y3 - y1*y3;

				if (b != 0)
				{
					// inflection
					t = a/b;
					if (t > 0+1e-6 && t < 1-1e-6)
						t_vals.push_back(t);

					// direction changes
					if (c > 0)
					{
						double d = sqrt(c);

						t = (a-d)/b;
						if (t > 0+1e-6 && t < 1-1e-6)
							t_vals.push_back(t);

						t = (a+d)/b;
						if (t > 0+1e-6 && t < 1-1e-6)
							t_vals.push_back(t);
					}
				}
			}

			// cut in x-direction
			{
				const double y0 = un->p[0][0];
				const double y1 = un->p[1][0];
				const double y2 = un->p[2][0];
				const double y3 = un->p[3][0];

				double a = y0 - 2*y1 + y2;
				double b = y0 - 3*y1 + 3*y2 - y3;
				double c = y1*y1 - y0*y2 - y1*y2 + y2*y2 + y0*y3 - y1*y3;

				if (b != 0)
				{
					// inflection
					t = a/b;
					if (t > 0+1e-6 && t < 1-1e-6)
						t_vals.push_back(t);

					// direction changes
					if (c > 0)
					{
						double d = sqrt(c);

						t = (a-d)/b;
						if (t > 0+1e-6 && t < 1-1e-6)
							t_vals.push_back(t);

						t = (a+d)/b;
						if (t > 0+1e-6 && t < 1-1e-6)
							t_vals.push_back(t);
					}
				}
			}

			//** actually do the cutting
			
			// if no cuts, just copy and quit
			assert(t_vals.s < 7);
			if (t_vals.s == 0)
			{
				*cp = *un;
				cp++;
				un++;
				continue;
			}

			// make sure the cuts are in order
			sort(t_vals.data, t_vals.data + t_vals.s);
			
			// cut it
			Line a, b, c = *un;
			double curr = 0;
			for (int i = 0; i < t_vals.s; i++)
			{
				double &t = t_vals[i];

				if (t - curr < 1e-6)
					continue; // skip duplicates

				cut_bez3(c, a, b, (t - curr) / (1 - curr));

				*cp = a;
				cp++;

				c = b;
				curr = t;
			}
			*cp = b;
			cp++;
		}
		else if (un->num_pts == 3)
		{
			// quadratic b-spline
			int n;
			Line s[2];

			// first check for a peak in y-direction
			int flip_y = un->p[2][1] < un->p[0][1] ? 2 : 0;
			const double y0 = un->p[flip_y][1];
			const double y1 = un->p[1][1];
			const double y2 = un->p[2-flip_y][1];

			if (y1 < y0 || y1 > y2)
			{
				double t = (y0 - y1)/(y0 - 2*y1 + y2);
				cut_bez2(*un, s[0], s[1], t);
				n = 2;
			}
			else
			{
				s[0] = *un;
				n = 1;
			}

			// now cut in the x-direction
			for (int i = 0; i < n; i++)
			{
				int flip_x = s[i].p[2][0] < s[i].p[0][0] ? 2 : 0;
				const double x0 = s[i].p[flip_x][0];
				const double x1 = s[i].p[1][0];
				const double x2 = s[i].p[2-flip_x][0];

				if (x1 < x0 || x1 > x2)
				{
					double t = (x0 - x1)/(x0 - 2*x1 + x2);
					cut_bez2(s[i], cp[0], cp[1], t);
					cp += 2;
				}
				else
				{
					*cp = s[i];
					cp++;
				}
			}
		}
		else
		{
			// a line is always monotonic, so copy it over
			*cp = *un;
			cp++;
		}

		un++;
	}

	cut.s = (int)(cp - cut.data);
#endif
}

void build_et(Array2D<double> &img, Array<Line> &lines_uncut)
{
	// cut all of the lines into monotonic pieces
#ifdef USE_ONLY_LINES
	Array<Line> &lines = lines_uncut;
#else
	Array<Line> lines;
	cut_monotonic(lines_uncut, lines);
#endif

	// init the scanline table
	for (int tid = 0; tid < NTHREADS; tid++)
	{
		ThreadData &td = tdata[tid];
		td.sl.heap.reserve(Scanline::MAX_CUT_LINES);
		if (img.s[0] > td.sl.lists.s)
		{
			td.sl.lists.resize(img.s[0]);
			
			// clear only if resizing, because it will be cleared during rasterization
			// later to prepare for next rasterization. This ensures that fewer zeros
			// have to be written over the long run.
			td.sl.lists = 0;
		}
	}

	// init the table
	et.lists.resize(img.s[1]);
	et.lists = 0;
	et.list_ends.resize(img.s[1]);
	et.list_ends = 0;
	et.heap.resize(lines.s); // (important!) the resize operation will not shrink the array

	et.start = img.s[1];
	et.end = 0;

	// insert curves to table
	for (int i = 0; i < lines.s; i++)
	{
		Line &l = lines[i];
		LineStep &c = et.heap[i];
		int start_y;

		// set curve points number
#ifdef USE_MIXED_CURVES
		c.num_pts = l.num_pts;
#endif

		if (c.num_pts == 2)
		{
			// calculate slope
			c.line.dy = (l.p[1][1] - l.p[0][1])/(l.p[1][0] - l.p[0][0]);
			c.line.dx = (l.p[1][0] - l.p[0][0])/(l.p[1][1] - l.p[0][1]);
		}
		else if (c.num_pts == 3)
		{
			// store mid point
			c.quad.mid_y = l.p[1];
		}
		else
		{
			// store mid points
			c.cube.mid_y[0] = l.p[1];
			c.cube.mid_y[1] = l.p[2];
		}

		// get edge range
		if (l.p[c.num_pts-1][1] < l.p[0][1])
		{
			if (c.num_pts == 4)
				swap(c.cube.mid_y[0], c.cube.mid_y[1]);

			c.flip_y = 1;
			start_y = (int)l.p[c.num_pts-1][1];
			c.end_y = (int)l.p[0][1];
			if (l.p[0][1] == floor(l.p[0][1]))
				c.end_y--;
			
			c.py[c.num_pts-1] = l.p[c.num_pts-1];
			c.high_y = l.p[0];
		}
		else
		{
			c.flip_y = 0;
			start_y = (int)l.p[0][1];
			c.end_y = (int)l.p[c.num_pts-1][1];
			if (l.p[c.num_pts-1][1] == floor(l.p[c.num_pts-1][1]))
				c.end_y--;
			
			c.py[c.num_pts-1] = l.p[0];
			c.high_y = l.p[c.num_pts-1];
		}

		// insert to table
		if (start_y <= c.end_y)
		{
			if (start_y < et.start)
				et.start = start_y;
			if (c.end_y > et.end)
				et.end = c.end_y;

			// stick in front of head
			c.next = et.lists[start_y];
			if (!et.lists[start_y]) // set as end if first element, because newer will be added in front
				et.list_ends[start_y] = &c;
			et.lists[start_y] = &c;
		}
	}
}

void *fix_ptr(void *base_dest, void *base_src, void *ptr_src)
{
	if (ptr_src == 0)
		return 0;
	else
		return (void *)((int)base_dest + (int)ptr_src - (int)base_src);
}

void copy_et(EdgeTable &dest, EdgeTable &src)
{
	dest.end = src.end;
	dest.start = src.start;

	dest.heap = src.heap;
	dest.lists.resize(src.lists.s);
	dest.list_ends.resize(src.lists.s);

	// fix pointers
	for (int il = 0; il < dest.lists.s; il++)
	{
		dest.lists[il] = (LineStep*)fix_ptr(&dest.heap[0], &src.heap[0], src.lists[il]);
		dest.list_ends[il] = (LineStep*)fix_ptr(&dest.heap[0], &src.heap[0], src.list_ends[il]);
	}

	for (int ih = 0; ih < src.heap.s; ih++)
	{
		dest.heap[ih].next = (LineStep*)fix_ptr(&dest.heap[0], &src.heap[0], src.heap[ih].next);
	}
}

void print_curves_y(LineStep *it, int curr_y)
{
	// print
	while (it)
	{
		if (it->num_pts == 4)
		{
			printf("{{%f,%f},{%f,%f},\n  {%f,%f},{%f,%f}},\n"
				,it->py[0][0],it->py[0][1] + curr_y
				,it->py[1][0],it->py[1][1] + curr_y
				,it->py[2][0],it->py[2][1] + curr_y
				,it->py[3][0],it->py[3][1] + curr_y);
		}
		else if (it->num_pts == 3)
		{
			printf("{{%f,%f},{%f,%f},{%f,%f}},\n"
				,it->py[0][0],it->py[0][1] + curr_y
				,it->py[1][0],it->py[1][1] + curr_y
				,it->py[2][0],it->py[2][1] + curr_y);
		}
		else if (it->num_pts == 2)
		{
			printf("{{%f,%f},{%f,%f}},\n"
				,it->py[0][0],it->py[0][1] + curr_y
				,it->py[1][0],it->py[1][1] + curr_y);
		}

		it = it->next;
	}
}

void rasterize(Array2D<double> &img, Array<Line> &lines)
{
	//freopen("output.txt", "wb", stdout);

	//** SERIAL: BUILD ACTIVE EDGE TABLE
	build_et(img, lines);

	//** PARALLEL: ITERATE OVER EDGE TABLE
	// threads render the image in multiple passes so that they don't interfere with eachother.
	// box filter is a special case, and can be done in one pass.
	
	//** determine how work should be broken up
	int bar_width = max(4, FILTER_SIZE - 1); // minimal width for 2 passes is W - 1. minimal width for W passes is 1
	int num_passes = 2;

	if (FILTER_SIZE == 1)
	{
		// box filter doesn't overlap, so only one pass is needed 
		bar_width = max(bar_width, 1);
		num_passes = 1;
	}

	int stride = bar_width * num_passes;
	
	int scan_start = et.start;
	int scan_end = et.end;

	//** run the rasterizer in parallel
	for (int pass = 0; pass < num_passes; pass++)
	{
		int bar_offset = pass * bar_width;
		int last_scanline[NTHREADS];

		// copy edge tables to each thread
		for (int tid = 0; tid < NTHREADS; tid++)
			tdata[tid].uninitialized = true;

		// process lines
#pragma omp parallel for schedule (dynamic,1)
		for (int bar = scan_start + bar_offset; bar <= scan_end; bar += stride)
		{
			int tid = omp_get_thread_num();
			ThreadData &td = tdata[tid];

			int range_start = bar;
			int range_end = min(scan_end, bar + bar_width-1);

			// copy edge table if first iteration for thread
			if (td.uninitialized)
			{
				td.uninitialized = false;
				td.ae_root = 0;
				last_scanline[tid] = scan_start;
				copy_et(td.etp, et);
			}

			// skip to beginning
			if (range_start > last_scanline[tid])
			{
				for (int iscan = last_scanline[tid]; iscan < range_start - 1; iscan++)
					skip_to_next_scanline(iscan, td);

				// set end point of $y$ curve segment and its end parameter value
				int iscan = range_start-1;
				skip_to_next_scanline(iscan, td);
				calc_end_points(td.etp.lists[iscan], iscan, last_scanline[tid], td);
			}

			// iterate over the scanlines
			for (int iscan = range_start; iscan <= range_end; iscan++)
			{
				update_ae_list(td.etp.lists[iscan], iscan, td);
				rasterize_line(img, iscan, td);

				// done with line, so clear it
				td.etp.lists[iscan] = 0;
			}

			last_scanline[tid] = range_end + 1;
		}

	}
}


